/**
  Finite Elements with OpenGL/GLUT
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2009-03-25 (Public Domain)
*/
#include "GL/glew.h" /* GL extension wrangler: http://glew.sourceforge.net/ */
#include "GL/glut.h" /* windows, mice, etc.  http://freeglut.sourceforge.net */
#include <stdlib.h> /* for "rand" */
#include <stdio.h> /* for printf */
#include "osl/vec4.h" /* 4D & 3D vector class */
#include "osl/mat2.h" /* 2D matrix and vector class */
using osl::mat2;
vec3 make3(const vec2 &v) {return vec3(v.x,v.y,0.0);}

#include "ogl/glsl.h" /* programmable shaders */
#include "ogl/glsl.cpp"

#include "ogl/minicam.h" /* small camera control functions */

#include <vector> /* std::vector container */

#ifndef M_PI
#define M_PI 3.14159265358979323
#endif

/* Random Float: Return a random number between 0 and 1 */
float rf(void) { return (rand()&0xfff)*(1.0/0xfff); }
/* Random Vector: return a random vector in 3D, all axes between 0 and 1 */
vec3 rv(void) {return vec3(rf(),rf(),rf());}

/* Return the wall clock time, in seconds */
double oglTimer(void) {return 0.001*glutGet(GLUT_ELAPSED_TIME);}
/* Check if an error has happened--if so, exit */
void oglCheck(const char *where) {
	int err=0;
	if (0!=(err=glGetError())) {
		fprintf(stderr,"OpenGL error %d at %s\n",err,where);
		exit(1);
	}
}

/* Counterclockwise test: return a positive value if, 
   from a's point of view, C is counterclockwise of B.
   If this returns zero, the points are colinear.
*/
double ccw(const vec2 &a,const vec2 &b,const vec2 &c) 
{
	vec2 B=b-a, C=c-a;
	return B.x*C.y-B.y*C.x;
}


class particle {
public:
	vec3 P; /* position */
	vec3 V; /* velocity */
	vec2 F; /* temporary summed-up force from all my triangles */
	vec2 outward; /* temporary summed-up outward-facing normal from my triangles */
	float mass; /* 1/3 * area of surrounding triangles * density */
	vec3 C; /* color */
	
	int collided; /* if not -1, the number of the triangle we're currently intersecting */
	vec2 collideF; /* collision force */
};

/* One finite-element triangle.  See Fedkiw et al 2004 paper for the basic idea. */
class triangle {
public:
	int corner[3]; /* array indexes of my particles */
	
	// Look up edge i: from corner i to corner i+1 
	vec2 edge(const particle *pts,int i) {
		return vec2(pts[corner[(i+1)%3]].P - pts[corner[i]].P);
	}
	
	// Normal of edge i, times the length of edge i
	vec2 edgenormal(const particle *pts,int i) {
		vec2 e=edge(pts,i);
		return vec2(-e.y,e.x);
	}
	
	// Return true if this particle is fully inside our triangle:
	bool contains(const particle *pts,int p) {
		return contains(pts,vec2(pts[p].P));
	}
	// Return true if this point is fully inside our triangle:
	bool contains(const particle *pts,vec2 testpoint) {
		/* check point against each edge */
		for (int d=0;d<3;d++) {
			vec2 s=vec2(pts[corner[d]].P); /* start of edge */
			vec2 e=vec2(pts[corner[(d+1)%3]].P); /* end of edge */
			if (ccw(s,e,testpoint)<+1.0e-9) // point is outside this edge
				return false;
		}
		return true;
	}
	
	// Body matrix: input barycentric triangle coords; output world coords
	mat2 body_matrix(const particle *pts) {
		return mat2(edge(pts,0),edge(pts,1));
	}
	
	mat2 rest_inverse; /* inverse of body matrix at startup */
	mat2 strain; /* strain matrix: how I've been squished */
	mat2 stress; /* stress matrix: how I push back (lbs/inch) */
	mat2 rotate; /* our element's rotation to get to world coordinates */
	float mass; /* mass of triangle */
	triangle() :stress(0.0) {}
	
	bool broked; 
	
	// Compute initial strain ("rest" configuration)
	void setup(particle *pts) {  
		broked=false;
		float density=1000.0; /* mass per unit area */
		float cross=ccw(vec2(0.0),edge(pts,0),edge(pts,1));
		if (cross<0) { /* triangle is oriented "wrong": flip inside out */
			std::swap(corner[1],corner[2]);
		}
		mass=0.5*fabs(cross)*density;
		for (int c=0;c<3;c++) /* account mass of triangle to its nodes */
			pts[corner[c]].mass+=(1.0/3.0)*mass;
		
		if (!contains(pts,vec2((1.0/3)*(pts[corner[0]].P+pts[corner[1]].P+pts[corner[2]].P))))
		{
			printf("Argh!  We don't even contain our own center!\n");
			exit(1);
		}
		
		mat2 rest=body_matrix(pts);
		rest_inverse=inverse(rest);
	}
	
	// Compute our stress, and apply forces to these particles
	void physics(particle *pts) {
		if (broked) return;
		
		strain=body_matrix(pts)*rest_inverse;
		
		// Remove the rotation part of my strain matrix
		vec2 sx=normalize(strain.x), sy=normalize(strain.y);
		//rotate=0.5*(mat2(sy.y,-sy.x, sy.x,sy.y)+mat2(sx.x,sx.y, -sx.y,sx.x));
		rotate=0.5*(mat2(sy.y,sy.x, -sy.x,sy.y)+mat2(sx.x,-sx.y, sx.y,sx.x));
		
		strain=strain*rotate;
		
		// Material constitutive model: see
		//    http://www.efunda.com/formulae/solid_mechanics/mat_mechanics/hooke_plane_stress.cfm
		float x_squish=strain.x.x-1.0;
		float y_squish=strain.y.y-1.0;
		float shear=strain.y.x+strain.x.y;
		float strength=0.3;
		if (fabs(x_squish)>strength || fabs(y_squish)>strength || fabs(shear)>strength)
			broked=true;
		
		shear*=0.25; /* shear strength is lower in real materials */
		
		
		float stiffness=5000.0;
		stress=stiffness*(
			mat2(x_squish,shear,  shear,y_squish)
		);
		
		// Apply the stress along each edge to get edge forces
		vec2 e[3],c[3]; // forces along each edge, corner
		e[0]=stress*edgenormal(pts,0);
		e[1]=stress*edgenormal(pts,1);
		e[2]=stress*edgenormal(pts,2);
		
		c[0]=e[2]+e[0];
		c[1]=e[0]+e[1];
		c[2]=e[1]+e[2];
		
		
		// Compute dissipation: in-element friction
		vec3 avgV=(1.0/3)*(pts[corner[0]].V+pts[corner[1]].V+pts[corner[2]].V);
		vec3 avgP=(1.0/3)*(pts[corner[0]].P+pts[corner[1]].P+pts[corner[2]].P);
		for (int i=0;i<3;i++) {
			particle &p=pts[corner[i]];
			vec2 relV=vec2(p.V-avgV);
			vec2 relP=normalize(vec2(p.P-avgP));
			p.outward+=relP; /* sum up outward-facing normal */
			relV=dot(relV,relP)*relP; /* project away rotational velocity */
			vec2 fF=-100.0*relV; /* beyond 10000, or less than 1.0: unstable*/
			
			// Write back our forces to the particle
			p.F+=fF+c[i];
		}
	}
	
	// Show our stress and strain configuration onscreen
	void draw(const particle *pts)
	{
		glBegin(GL_LINES);
		vec2 center=vec2((1.0/3)*(
			pts[corner[0]].P+
			pts[corner[1]].P+
			pts[corner[2]].P
		));
	
		for (float ang=0;ang<2*M_PI-0.001;ang+=M_PI*0.5) 
		{
			float angdark=0.3+ang*0.1;
			vec2 dir(cos(ang),sin(ang));
			
			glColor3f(0,1*angdark,0); /* strain in green */
			glVertex2fv(center);
			glVertex2fv(center+strain*dir*0.1);
			
			glColor3f(1*angdark,0,0); /* stress in red */
			glVertex2fv(center);
			glVertex2fv(center+stress*dir*0.01);
			
			glColor3f(1*angdark,1*angdark,0); /* rotation in yellow */
			glVertex2fv(center);
			glVertex2fv(center+rotate*dir*0.1);
		}
		glEnd();
	}
};

using namespace std;

/**
  "Narrow Phase" collision detection & response.
  
  Checks if point "pt" intersects this triangle, 
  and if so, pushes back on both of 'em.
*/
void collide(particle *pts,triangle &t,int pt)
{
	if (t.contains(pts,pt)) 
	{ /* we had a collision */
		vec2 dir=normalize(pts[pt].outward); /* normal force */
		
		float x=1.0e30; /* penetration distance */
		for (int e=0;e<3;e++) {
			vec2 A=vec2(pts[t.corner[e]].P);
			vec2 B=vec2(pts[t.corner[(e+1)%3]].P);
			vec2 I=vec2(pts[pt].P);
			float mx=ccw(A,B,I)/length(A-I);
			x=std::min(x,mx);
		}

		vec2 F=-1.0e3*x*dir; /* normal force */
		pts[pt].collideF=F; /* display only */

		pts[pt].F+=F;/* push back on point */
		for (int c=0;c<3;c++) /* equal and opposite reaction on triangle */
			pts[t.corner[c]].F-=(1.0/3.0)*F;

		pts[pt].collided=1;
	}
}

/**
 "Broad Phase" collision detection acceleration data structure:
 keeps a grid of particles, and intersects triangles with only 
 the particles in the grid cells touched by the triangle.
*/
class collidegrid {
	/* For each x,y, a list of nodes in that grid cell */
	std::vector< std::vector< std::vector<int> > > grid;
	int res; /* number of grid cells along each axis */
	float gridFmWorld;  /* convert world coordinates into grid cell index */
	int grid_y(const vec3 &p) {
		return (int)(p.y*gridFmWorld);
	}
	int grid_x(const vec3 &p) {
		return (int)(p.x*gridFmWorld);
	}
public:
	collidegrid(int res_) 
		:res(res_), gridFmWorld(res)
	{
		for (int y=0;y<res;y++) {
			std::vector< std::vector<int> > row;
			for (int x=0;x<res;x++) row.push_back(std::vector<int>());
			grid.push_back(row);
		}
	}
	
	/* Add particles to my grid */
	void sift(const particle *pts,int npts) {
		/* Reset each grid cell's list of particles */
		for (int y=0;y<res;y++) for (int x=0;x<res;x++)
			grid[y][x].clear();
		
		/* Add particles to my grid */
		for (int p=0;p<npts;p++) {
			int y=grid_y(pts[p].P);
			int x=grid_x(pts[p].P);
			if (x>=0 && x<res && y>=0 && y<res) // bounds check
				grid[y][x].push_back(p);
		}
	}
	
	/* If this triangle collides with any points, push 'em back */
	void hit(particle *pts,triangle &t){
		/* Find the grid cells I touch */
		int maxX=0,maxY=0;
		int minX=res-1,minY=res-1;
		for (int p=0;p<3;p++) {
			int y=grid_y(pts[t.corner[p]].P);
			int x=grid_x(pts[t.corner[p]].P);
			maxX=max(maxX,x); maxY=max(maxY,y); 
			minX=min(minX,x); minY=min(minY,y);
		}
		if (minY<0 || minX<0) return; /* NaN: skip triangle */
		if (maxX>res-1) maxX=res-1; // bounds check
		if (maxY>res-1) maxY=res-1;

		/* Check the points stored in each of my grid cells */		
		for (int y=minY;y<=maxY;y++) 
		for (int x=minX;x<=maxX;x++) 
		{
			const std::vector<int> &list=grid[y][x];
			for (unsigned int p=0;p<list.size();p++) {
				collide(pts,t,list[p]);
			}
		} 
	}
};


void display(void) /* Draw everything in our window */
{
	unsigned int i;
	oglCheck("before display function");
	static double elapsed=0.0; /* last elapsed display time (seconds) */
	static float dt=0.001, display_dt=0.0;
	static float sim_time=0.0;
	float big=30; /* size of the ground */
	float floor=0.0; /* Z height of floor */
	glFinish(); double start_t=oglTimer();
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
/* Particle state storage */
	static bool first_time=true; /* for initial bc's! */
	static std::vector<particle> pts;
	static std::vector<triangle> triangles;
	if (first_time) { /* First time: initialize points */
		static int count=0;
		srand(6);
		std::ifstream nf("mesh/image.1.node");  std::string crap;
		
		if (false || !nf) { /* hardcoded mesh: one triangle */
			for (int i=0;i<9;i++) {
				particle p;
				p.P=0.3*rv()+0.1*vec3(0,i,0); 
				p.P.z=0.0001;
				p.V=0.03*(rv()-vec3(0.5)); 
				p.mass=1.0;
				p.C=rv();
				pts.push_back(p);

				if ((i+1)%3==0) {
					triangle t;
					t.corner[0]=i-2;
					t.corner[1]=i-1;
					t.corner[2]=i;
					t.setup(&pts[0]);
					triangles.push_back(t);
				}
			}
		} else { /* read mesh from files */
			unsigned int nnode=0; nf>>nnode; getline(nf,crap); /* first line: # nodes */
			for (i=0;i<nnode;i++) {
				int nodeno; float x,y;
				nf>>nodeno>>x>>y; std::getline(nf,crap); /* node lines */
				//if (nodeno==1) continue; /* first node is not connected to anything (!?) */
				particle p;
				p.P=0.15*vec3(x,y,0)+vec3(0.5); 
				p.P.z=0.0001;
				p.V=vec3(0.0);
				if (p.P.x>0.5) p.V.y+=0.1; // initial rotation 
				p.V.z=0.0;
				p.mass=0.1; /* fake base mass; mass is also added per triangle */
				p.C=rv();
				pts.push_back(p);
			}

			std::ifstream tf("mesh/image.1.ele"); 
			unsigned int ntri=0;
			tf>>ntri; std::getline(tf,crap);
			for (i=0;i<ntri;i++) {
				triangle t;
				int trino, a,b,c;
				tf>>trino>>a>>b>>c; std::getline(tf,crap);
				t.corner[0]=a-1; /* -1 for one-based; -2 if first node is gone */
				t.corner[1]=b-1;
				t.corner[2]=c-1;
				t.setup(&pts[0]);
				triangles.push_back(t);
			}
			
			for (int p=0;p<pts.size();p++) printf("Mass of %i: %f\n",p,pts[p].mass);
		}
		
		first_time=true;
	}
	
	
/* Main camera setup */
	glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	glClearColor(0.3,0.5,0.7,1.0); /* clear to a nice blue color */
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70.0, /* degree FOV */
		glutGet(GLUT_WINDOW_WIDTH)/float(glutGet(GLUT_WINDOW_HEIGHT)),
		0.001,1000.0 /* Z range */
	);
	oglCameraLookAt(1.0);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// Draw in a ground plane
	glDisable(GL_DEPTH_TEST); /* don't read/write from depth buffer */
	//glDepthMask(GL_FALSE); /* don't write to depth buffer (no Z fighting!) */
	
	glBegin(GL_QUADS);
	glColor4f(0.3,0.3,0.3,1.0); /* light, opaque brown */
	glVertex3f(-big,-big,floor);
	glVertex3f(+big,-big,floor);
	glVertex3f(+big,+big,floor);
	glVertex3f(-big,+big,floor);
	glEnd();

	// Draw in a 1-meter grid on the ground
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_BLEND);
	glLineWidth(0.4);
	glBegin(GL_LINES);
	glColor4f(0.0f,0.0f,0.0f,0.3f);
	for (float v=-10;v<=10;v+=0.1) {
		glVertex3f(-big,v,floor); /* X lines */
		glVertex3f(+big,v,floor);
		glVertex3f(v,-big,floor); /* Y lines */
		glVertex3f(v,+big,floor);
	}
	glEnd();
	
	glDepthMask(GL_TRUE); /* depth writes back on */
	
	
	/* Subcycle, so the physics runs at the right stepsize, 
	   yet still keeps up with the display. */
	double wall_time=oglTimer();
	double target_time=wall_time*1.0; // <- constant for slow-motion 
	if (key_down[(int)' ']) sim_time=target_time+1.0e-3; /* spacebar == stop physics */
	while (sim_time<target_time)
	{ /* take one physics step */
		if (oglTimer()>wall_time+0.1) break; /* minimum of 10Hz animation */
		sim_time+=dt;
		
		float energy=0.0;
		vec2 outward_init(0.0,1.0e-5);
		/* zero net force */
		for (i=0;i<pts.size();i++) {
			pts[i].F=vec2(0.0); 
			pts[i].collided=-1;
			pts[i].outward=outward_init;
		}
		
		/* Apply user interface force to the first particle */
		int keypart=pts.size()/2;
		float keyforce=30.0;
		if (key_down[(int)'x']) pts[keypart].F.x+=keyforce;
		if (key_down[(int)'X']) pts[keypart].F.x-=keyforce;
		if (key_down[(int)'y']) pts[keypart].F.y+=keyforce;
		if (key_down[(int)'Y']) pts[keypart].F.y-=keyforce;
		//if (key_down[(int)'z']) pts[keypart].F.z+=50.0;
		
		/* Compute the triangles' strain and stress to find the net force */
		for (i=0;i<triangles.size();i++) {
			triangle &t=triangles[i];
			t.physics(&pts[0]);
		}
		
		/* collision detection force computation */
		static collidegrid grid(100);
		grid.sift(&pts[0],pts.size());
		
		for (unsigned int tri=0;tri<triangles.size();tri++) 
		if (triangles[tri].broked==false)
			grid.hit(&pts[0],triangles[tri]);
		
		/* physics the particles */
		for (i=0;i<pts.size();i++) {
			particle &p=pts[i];
			
			/* Unconnected points--move off to corner */
			if (pts[i].outward==outward_init) p.P=p.V=vec3(0.0);

		/* Apply physics */
			vec3 gA(0.0,-0.1,0.0); /* gravity acceleration */

			//vec3 fF=-0.5*p.V;

			vec3 A=(vec3(p.F.x,p.F.y,0))/p.mass +gA; /* total acceleration */
			
			energy+=p.mass*dot(p.V,p.V);
			
			p.P+=dt*p.V+0.5*dt*dt*A; /* move position by velocity */
			p.V+=dt*A; /* accelerate particles downwards */
			p.V.z=0.0;
			
			/* Bounce off all six sides of cube */
			for (int dir=-1;dir<=1;dir+=2)
			for (int axis=0;axis<3;axis++) {
				float loc=(p.P[axis]-0.5)*2.0;
				if (loc*dir>1.0 && p.V[axis]*dir>0.0) { /* bounce off this axis */
					p.V[axis]*=-1.0;
				}
			}
		}
		
		if (energy>1.0e3) printf("t=%.3f, energy=%f\n",sim_time,energy);
		if (energy!=energy) exit(1); /* NaN energy: unstable */

		first_time=false;
	}
	
/* Draw axes */
	oglCameraAxes();

/* Draw triangles */
	glLineWidth(1.0);
	for (i=0;i<triangles.size();i++) {
		triangle &t=triangles[i];
		if (t.broked) continue;
		
		glColor3f(0,0,0); /* black outline */
		glBegin(GL_LINE_LOOP);
		for (int j=0;j<3;j++) glVertex2fv(pts[t.corner[j]].P);
		glEnd();
		
		if (i<0) /* draw stress on selected triangles */
			t.draw(&pts[0]);
	}
	
/* Draw our points */
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_POINT_SMOOTH);
	glPointSize(3.0);
	
	glBegin(GL_POINTS);
	for (i=0;i<pts.size();i++) {
		particle &p=pts[i];
	/* Draw the particle */
		if (p.collided>-1) {
			glEnd(); /* GL_POINTS, so we can draw lines */
			
			glLineWidth(4.0);
			glBegin(GL_LINES);
			glColor3f(1,0,0);
			glVertex3fv(p.P);
			glVertex3fv(p.P+0.001*vec3(p.collideF.x,p.collideF.y,0.0));
			glEnd();
			
			glBegin(GL_POINTS);
			glColor3f(1,0,0);
		}
		else glColor3fv(p.C);
		glVertex3fv(p.P);
	}
	glEnd();	
	
	glutSwapBuffers(); /* display everything we've rendered so far */
	glutPostRedisplay(); /* continual animation */
	
	// Print frames per second
	glFinish(); elapsed=oglTimer()-start_t; /* stop the stopwatch */
	static double last_print=0.0; /* Periodically print out the time */
	if (start_t>last_print+0.1) {
		display_dt=elapsed;
		printf("%.1f frames/sec, %.2f ms/frame\n",
			1.0/elapsed,1000.0*elapsed);
		last_print=start_t;
	}
	oglCheck("after display function");
}

int main(int argc,char **argv) {
	camera=vec3(0.5,0.5,1.0);
	camera_orient.y=vec3(0,1,0); camera_orient.z=vec3(0,0,1);

	glutInit(&argc,argv);  /* set up GLUT */
	glutInitWindowPosition(0,60);
	glutInitWindowSize(1000,700);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	glutCreateWindow("FEM + collision demo (try WASD and XYZ keys!)"); /* make a window */
	glewInit(); /* only AFTER create window! */
	glutDisplayFunc(display); /* here's how to draw the window */
	oglCameraInit();
	glutMainLoop(); /* start drawing */
}
