/* SVG file parser
Dr. Orion Lawlor, olawlor@acm.org, 2008-08-18 (Public Domain)
*/
#include <stdio.h>
#include <stdlib.h>
#include "svg_parse.h"
#include "tinyxml.h" /* see http://tinyxml.sourceforge.net/ */
#include "tinyxml.cpp" 
#include "tinyxmlparser.cpp"

void die(const char *str,const char *str2="") {printf(str,str2);exit(1);}

/* Parse an SVG transform description into a 2D matrix */
mat2 parse_matrix(const char *svgMat) {
	mat2 m(1.0);
	if (2==sscanf(svgMat,"translate(%f,%f)",&m(0,2),&m(1,2))) 
		return m;
	if (6==sscanf(svgMat,"matrix(%f,%f,%f,%f,%f,%f)",
		&m(0,0),&m(1,0),  &m(0,1),&m(1,1),  &m(0,2),&m(1,2)
		)) return m;
	die("Can't parse matrix %s\n",svgMat);
	return m;
}

/* Parse an SVG path description into a series of vectors.
	E.g.:  "M 0,3  L 2,3  L 4,5  z" will turn into 3 vectors.
*/
std::vector<vec2> parse_path(const char *path) {
	std::vector<vec2> ret;
	int n;
	char c;
	float x,y;
	while (3<=sscanf(path," %c %f,%f %n",
		&c,&x,&y,&n)) 
	{
		ret.push_back(vec2(x,y));
		path+=n; /* skip over n bytes */
	}
	return ret;
}

/* Parse out the useful stuff in an SVG style description.
	E.g., "fill:none;foo:bar;crap:OK" 
*/
svg_style parse_style(const char *str_style) {
	svg_style style; style.width=0; style.opacity=0;
	std::string str=str_style;
	size_t s=0;
	do {
		size_t c=str.find(":",s); /* points to middle */
		size_t e=str.find(";",c); /* points to end (or npos) */
		std::string kw(str,s,c-s);
		std::string val(str,c+1,e-(c+1));
		//printf("	'%s' = '%s'\n",kw.c_str(),val.c_str());
		
		if (kw=="stroke-width") 
			style.width=atof(val.c_str());
		if (kw=="stroke-opacity")
			style.opacity=atof(val.c_str());
		if (kw=="fill-opacity") {
			style.opacity=atof(val.c_str());
			break; /* stop right now-- don't parse line attributes! */
		}
		
		s=e+1; /* start of next is end of this */
	} while (s!=std::string::npos+1);
	return style;
}

/* Allows cool node^"field" syntax to be used on an XML file. */
inline TiXmlElement& operator^(TiXmlNode &parent,const char *field) {
	TiXmlNode* n=parent.FirstChild(field);
	if (!n) die("Error looking up XML child node: %s!",field);
	TiXmlElement *e=n->ToElement();
	if (!e) die("Error making element from XML child node: %s!",field);
	return *e;
}

void svg_parse(const char *fName,svg_consumer &dest) {
	TiXmlDocument doc(fName);
	if (!doc.LoadFile()) die("Error loading file: %s!",doc.ErrorDesc());
	TiXmlElement &main=doc^"svg"^"g";
	for ( /* Loop over the elements of the SVG file */
		TiXmlElement* ep=main.FirstChildElement();
		ep!=NULL;
		ep=ep->NextSiblingElement()) 
	{
		TiXmlElement &e=*ep;
		std::string v=e.ValueStr();
		//print("Entity '%s'\n",e);
		if (v=="use") {
			dest.svg_use(
				e.Attribute("xlink:href"),
				parse_matrix(e.Attribute("transform"))
			);
		}
		else if (v=="path") {
			dest.svg_path(
				//e.Attribute("id"),
				parse_path(e.Attribute("d")),
				parse_style(e.Attribute("style"))
			);
		}
		/*
		else if (v=="text") {
			print("	Text '%s'\n",(e^"tspan").GetText());
		}*/
	}
}

#if 0
/******* Test harness--print out SVG file contents *********/
void print(const char *str,const TiXmlElement &e) {
	printf(str,e.Value());
}
void print(const char *str,const mat2 &m) {
	printf("%s: \n"
		"		%f %f %f\n"
		"		%f %f %f\n",
		str,
		m(0,0), m(0,1), m(0,2), 
		m(1,0), m(1,1), m(1,2)
	);
}
void print(const char *str,const std::vector<vec2> &p) {
	printf("%s: \n",str);
	for (unsigned int i=0;i<p.size();i++)
		printf("	%f,%f\n",p[i].x,p[i].y);
}
void print(const char *str,const svg_style &s) {
	printf("%s: width %f   opacity %f\n",str,s.width,s.opacity);
}
void print(const char *str,const char *e) {
	printf(str,e);
}
int main() {
	class svg_printer : public svg_consumer {
	public:
		virtual void svg_use(const char *what,const mat2 &where) {
			print("	Using %s\n",what);
			print("	At",where);
		}
		virtual void svg_path(const std::vector<vec2> &pts,const svg_style &style) {
			print("	Path",pts);
			print("	Style",style);
		}
	};
	svg_printer dest;
	svg_parse("test.svg",dest);
}
#endif
